@@ -1,9 +1,4 @@
-{-
- -      ``Control/Monad/Loops''
- -}
-{-# LANGUAGE
-        CPP
-  #-}
+{-# LANGUAGE CPP #-}
 
 -- |A collection of loop operators for use in monads (mostly in stateful ones).
 -- 
@@ -21,9 +16,6 @@
 
 module Control.Monad.Loops
         ( module Control.Monad.Loops
-#ifdef useSTM
-        , module Control.Monad.Loops.STM
-#endif
         ) where
 
 import Control.Monad
@@ -35,12 +27,6 @@ import Control.Concurrent
 #define SomeException Exception
 #endif
 
-#ifdef useSTM
-import Control.Monad.Loops.STM
-#endif
-
-import Data.Maybe (listToMaybe)
-
 -- possibly-useful addition? :
 -- concatMapM :: (Monad m, Traversable f, Monoid w) => (a -> m w) -> (f a) -> m w
 
@@ -90,7 +76,7 @@ forkMapM__ f xs = do
                         -- in base >=4, need to nail down the type of 'handle'
                         let handleAny :: (SomeException -> IO a) -> IO a -> IO a
                             handleAny = handle
-                        handleAny (\e -> return ()) $ do
+                        handleAny (\_ -> return ()) $ do
                                 f x
                                 return ()
                         putMVar mvar ()
@@ -137,12 +123,14 @@ whileM_ p f = go
 -- |Execute an action repeatedly until its result fails to satisfy a predicate,
 -- and return that result (discarding all others).
 iterateWhile :: Monad m => (a -> Bool) -> m a -> m a
-iterateWhile p x = go
-    where go = do
-                y <- x
-                if p y
-                    then go
-                    else return y
+iterateWhile p = iterateUntil (not . p)
+
+{-# SPECIALIZE iterateM_ :: (a -> IO a) -> a -> IO b #-}
+-- |Execute an action forever, feeding the result of each execution as the
+-- input to the next.
+iterateM_ :: Monad m => (a -> m a) -> a -> m b
+iterateM_ f = g
+    where g x = f x >>= g
 
 {-# SPECIALIZE untilM  :: IO a -> IO Bool -> IO [a] #-}
 {-# SPECIALIZE untilM' :: Monad m => m a -> m Bool -> m [a] #-}
@@ -152,6 +140,7 @@ iterateWhile p x = go
 infixr 0 `untilM`
 infixr 0 `untilM'`
 infixr 0 `untilM_`
+infixr 0 `iterateUntilM`
 
 -- |Execute an action repeatedly until the condition expression returns True.
 -- The condition is evaluated after the loop body.  Collects results into a list.
@@ -175,20 +164,23 @@ f `untilM'` p = do
 untilM_ :: (Monad m) => m a -> m Bool -> m ()
 f `untilM_` p = f >> whileM_ (liftM not p) f
 
-{-# SPECIALIZE whileJust  :: IO (Maybe a) -> (a -> IO b) -> IO [b] #-}
-{-# SPECIALIZE whileJust' :: Monad m => m (Maybe a) -> (a -> m b) -> m [b] #-}
-{-# SPECIALIZE whileJust' :: IO (Maybe a) -> (a -> IO b) -> IO [b] #-}
-{-# SPECIALIZE whileJust_ :: IO (Maybe a) -> (a -> IO b) -> IO () #-}
+
+-- | Analogue of @('Prelude.until')@
+-- Yields the result of applying f until p holds.
+iterateUntilM :: (Monad m) => (a -> Bool) -> (a -> m a) -> a -> m a
+iterateUntilM p f v 
+    | p v       = return v
+    | otherwise = f v >>= iterateUntilM p f
 
 -- |Execute an action repeatedly until its result satisfies a predicate,
 -- and return that result (discarding all others).
 iterateUntil :: Monad m => (a -> Bool) -> m a -> m a
-iterateUntil p x = go
-    where go = do
-            y <- x
-            if p y
-                then return y
-                else go
+iterateUntil p x = x >>= iterateUntilM p (const x)
+
+{-# SPECIALIZE whileJust  :: IO (Maybe a) -> (a -> IO b) -> IO [b] #-}
+{-# SPECIALIZE whileJust' :: Monad m => m (Maybe a) -> (a -> m b) -> m [b] #-}
+{-# SPECIALIZE whileJust' :: IO (Maybe a) -> (a -> IO b) -> IO [b] #-}
+{-# SPECIALIZE whileJust_ :: IO (Maybe a) -> (a -> IO b) -> IO () #-}
 
 -- |As long as the supplied "Maybe" expression returns "Just _", the loop
 -- body will be called and passed the value contained in the 'Just'.  Results
@@ -291,13 +283,13 @@ unfoldrM = unfoldrM'
 -- twist.  Rather than returning a list, it returns any MonadPlus type of your
 -- choice.
 unfoldrM' :: (Monad m, MonadPlus f) => (a -> m (Maybe (b,a))) -> a -> m (f b)
-unfoldrM' f z = go
-    where go = do
+unfoldrM' f = go
+    where go z = do
             x <- f z
             case x of
                 Nothing         -> return mzero
-                Just (x, z)     -> do
-                        xs <- go
+                Just (x, z')    -> do
+                        xs <- go z'
                         return (return x `mplus` xs)
 
 {-# SPECIALIZE concatM :: [a -> IO a] -> (a -> IO a) #-}
@@ -336,7 +328,7 @@ orM (p:ps)      = do
 -- value presented against each predicate in turn until one passes, then
 -- returns True without any further processing.  If none passes, returns False.
 anyPM :: (Monad m) => [a -> m Bool] -> (a -> m Bool)
-anyPM []     x = return False
+anyPM []     _ = return False
 anyPM (p:ps) x = do
         q <- p x
         if q
@@ -347,7 +339,7 @@ anyPM (p:ps) x = do
 -- presented against each predicate in turn until one fails, then returns False.
 -- if none fail, returns True.
 allPM :: (Monad m) => [a -> m Bool] -> (a -> m Bool)
-allPM []     x = return True
+allPM []     _ = return True
 allPM (p:ps) x = do
         q <- p x
         if q
@@ -359,7 +351,7 @@ allPM (p:ps) x = do
 
 -- |short-circuit 'any' with a \"monadic predicate\".
 anyM :: (Monad m) => (a -> m Bool) -> [a] -> m Bool
-anyM p []       = return False
+anyM _ []       = return False
 anyM p (x:xs)   = do
         q <- p x
         if q
@@ -368,7 +360,7 @@ anyM p (x:xs)   = do
 
 -- |short-circuit 'all' with a \"monadic predicate\".
 allM :: (Monad m) => (a -> m Bool) -> [a] -> m Bool
-allM p []       = return True
+allM _ []       = return True
 allM p (x:xs)   = do
         q <- p x
         if q
@@ -376,7 +368,7 @@ allM p (x:xs)   = do
                 else return False
 
 dropWhileM :: (Monad m) => (a -> m Bool) -> [a] -> m [a]
-dropWhileM p []     = return []
+dropWhileM _ []     = return []
 dropWhileM p (x:xs) = do
         q <- p x
         if q
@@ -392,7 +384,7 @@ trimM p xs = do
 
 -- |return the first value from a list, if any, satisfying the given predicate.
 firstM :: (Monad m) => (a -> m Bool) -> [a] -> m (Maybe a)
-firstM p [] = return Nothing
+firstM _ [] = return Nothing
 firstM p (x:xs) = do
         q <- p x
         if q
